"""Module create nlp labelling project in labelbox."""
import os
import warnings
from typing import Optional, Union
from pathlib import Path
import json
import pandas as pd
import boto3
import s3fs
from labelbox import Client  # , Dataset,  DataRow,

warnings.filterwarnings("ignore")


s3 = s3fs.S3FileSystem()


def pull_files_s3(prefix: str, file_type: str, s3_bucket, delimiter='/') -> str:
    """pull_files_s3

    Args:
        prefix (str): S3 folder path name without the bucket part
        file_type (str): type of the data file
        s3_bucket (_type_): S3 bucket
        delimiter (str, optional): Defaults to '/'.

    Yields:
        Iterator[str]: _description_
    """
    prefix = prefix[1:] if prefix.startswith(delimiter) else prefix
    yield from (_.key for _ in s3_bucket.objects.filter(Prefix=prefix) if file_type in _.key)


def create_labeling_project(labelbox_api_key: str,
                            data: Union[str, Path, pd.DataFrame], annotation_col: str, id_col: str,
                            s3_bucket_name: str, dataset_name: str, project_name: str, project_desc: Optional[str] = '',
                            prefix: Optional[str] = None,
                            temp_data_path: Optional[Union[str, Path]] = None) -> dict:
    """create_labeling_project creates labelling project. dataset and attaches data to project from dxi S3 bucket.

    Args:
        labelbox_api_key (str): your labelbox api_key
        data (Union[str, Path, pd.DataFrame]): data as a dataframe or file path.
        annotation_col (str): the column that we need to annotate in the labelling project.
        id_col (str): the unique identifier column of the dataframe.
        dataset_name (str): name you want to assign to your dataset on labelbox. (json will be added to end)
        project_name (str): name you want to assign to your project on labelbox.
        project_desc (Optional[str]): task description of the project.
        s3_bucket_name (str, optional): S3 bucket that is integrated with labelbox. Defaults to 'test-labelbox-edelman'.
        prefix (Optional[str], optional): folder path to store text files in the S3 bucket. Defaults to None.
        temp_data_path (Optional[Union[str, Path]], optional): file path where text files are stored temporarily. Defaults to None.

    Returns:
        dict: _description_
    """
    # folder to store and sync temporary text files to S3
    if temp_data_path is None:
        temp_data_path = Path(Path.cwd()/'temp_text_file_storage')
        temp_data_path.mkdir(exist_ok=True, parents=True)

    for _, row in data.iterrows():
        with open(temp_data_path/f'{row[id_col]}.txt', 'w') as file:
            file.write(row[annotation_col])

    sync_command = f"aws s3 sync {str(temp_data_path)}  s3://{s3_bucket_name}/{prefix} --quiet --acl public-read"
    os.system(command=sync_command)

    s3_bucket = boto3.resource('s3').Bucket(s3_bucket_name)
    s3_files = list(pull_files_s3(
        prefix=prefix, file_type='txt', s3_bucket=s3_bucket))

    client = Client(api_key=labelbox_api_key)

    dataset_name += ".json"
    # creates an empty dataset
    dataset = client.create_dataset(name=dataset_name)

    # creates an empty project
    project = client.create_project(
        name=project_name,
        description=project_desc)

    # Attach dataset
    project.datasets.connect(dataset)
    my_data_rows = [
        {
            "row_data": f'https://{s3_bucket_name}.s3.amazonaws.com/{file}',
            "external_id": file.split('/')[-1].split('.txt')[0]
        } for file in s3_files
    ]

    s3_file_path = f"{s3_bucket_name}/{dataset_name}"
    with s3.open(s3_file_path, 'w') as file:
        json.dump(my_data_rows, file)

    task = dataset.create_data_rows(my_data_rows)
    task.wait_till_done()
    print(f'sync labelbox {dataset_name} with text files on S3: ', task.status)

    files = temp_data_path.glob('*')
    _ = [file.unlink() for file in files]
    temp_data_path.rmdir()

    return {'created_project': project_name,
            'project_uid': project.uid,
            'attached_datset': dataset_name,
            'dataset_uid': dataset.uid,
            'data_loaded_from': f"s3://{s3_bucket_name}/{prefix}",
            'aws_url_reference_json': f"s3://{s3_file_path}"
            }



"""Module to extract relevant sentences and their scores."""
from typing import Union
import re
import gc
from pathlib import Path
import pandas as pd
from .relevancy_utils import nlp_lg, split_clean_paragraphs, split_clean_sentences, compute_relevancy_score_for_content, compute_relevancy_score_for_title


def get_relevant_sentences_from_lexis(
        df: Union[str, Path, pd.DataFrame],
        file_type: str = 'parquet',
        multiprocess: bool = False,
        n_process: int = 1,
        batch_size: int = 1000) -> pd.DataFrame:

    read_fn = {
        'parquet': pd.read_parquet, 'feather': pd.read_feather,
        'csv': pd.read_csv, 'json': pd.read_json
    }

    if not isinstance(df, pd.DataFrame):
        if file_type == 'json':
            df = read_fn[file_type](df, lines=True)
        df = read_fn[file_type](df)

    df['article_word_count'] = df.content.apply(lambda x: len(x.split()))

    df['brand_in_title'] = df.apply(
        lambda x: 1 if x.brand_2.lower() in x.title.lower() or x.brand.lower() in x.title.lower() else 0, axis=1)

    df['paragraphs'] = df.apply(split_clean_paragraphs, axis=1)

    df_exploded = df.explode('paragraphs')
    df_exploded = df_exploded[~df_exploded.paragraphs.isna(
    )]
    df_exploded.paragraphs = df_exploded.paragraphs.apply(
        lambda x: re.sub('\n\n', ' : ', re.sub(' +', ' ', x)))
    df_exploded['brand_in_paragraph'] = df_exploded.apply(
        lambda x: 1 if x.brand_2.lower() in x.paragraphs.lower(
        ) or x.brand.lower() in x.paragraphs.lower() else 0,
        axis=1)

    df_brand_in_paragraph = df_exploded[
        [
            'brand_article_id', 'brand', 'brand_2', 'paragraphs'
        ]
    ][df_exploded.brand_in_paragraph == 1]
    df_brand_in_paragraph.drop_duplicates(
        subset=['brand', 'brand_2', 'brand_article_id', 'paragraphs'], inplace=True)
    df_brand_in_paragraph.reset_index(inplace=True, drop=True)

    del df_exploded
    gc.collect()

    paragraph_list = df_brand_in_paragraph.paragraphs.tolist()

    if multiprocess:
        paragraph_doc = list(nlp_lg.pipe(
            paragraph_list, n_process=n_process, batch_size=batch_size))
    else:
        paragraph_doc = list(nlp_lg.pipe(paragraph_list))

    df_brand_in_paragraph['paragraph_doc'] = paragraph_doc

    del paragraph_doc, paragraph_list
    gc.collect()

    df_brand_in_paragraph['sentence'] = df_brand_in_paragraph.apply(
        lambda x: split_clean_sentences(x.paragraph_doc, x.brand, x.brand_2), axis=1)
    df_brand_in_paragraph = df_brand_in_paragraph.explode('sentence')
    df_brand_in_paragraph = df_brand_in_paragraph[~df_brand_in_paragraph.sentence.isna(
    )]
    df_brand_in_paragraph.drop_duplicates(
        subset=['brand', 'brand_2', 'sentence', 'brand_article_id'], inplace=True)
    df_brand_in_paragraph.reset_index(inplace=True, drop=True)

    df_brand_in_paragraph['brand_article_relevancy_score'] = df_brand_in_paragraph.apply(
        lambda x: compute_relevancy_score_for_content(x.sentence, x.brand, x.brand_2), axis=1)

    return df_brand_in_paragraph



"""Utility module to enable all relevancy apis."""
from typing import Iterable, Generator
import re
# import gc
import spacy
from spacy import Language
# import pandas as pd

nlp_lg = spacy.load('en_core_web_lg', exclude=['lemmatizer', 'ner'])
# nlp_trf = spacy.load('en_core_web_trf', exclude=['lemmatizer', 'ner'])


@Language.component('set_custom_boundaries')
def set_custom_boundaries(doc: spacy.tokens.doc.Doc) -> spacy.tokens.doc.Doc:
    """set_custom_boundaries adds custom characters for sentence segmentation.
    Args:
        doc (spacy.tokens.doc.Doc): spacy language document.
    Returns:
        spacy.tokens.doc.Doc: spacy language document.
    """
    # Adds support to use `` as the delimiter for sentence detection
    for delim in ['|', ':', ':::']:
        for token in doc[:-1]:
            if token.text == delim:
                doc[token.i+1].is_sent_start = True
    return doc


# nlp_trf.add_pipe('set_custom_boundaries', before='parser')
nlp_lg.add_pipe('set_custom_boundaries', before='parser')


def flatten(nest: Iterable) -> Generator:
    """Flatten takes a nested iterator and yields from a flattened generator.
    Args:
        nest (Iterable): nested iterator
    Yields:
        Iterator[Generator]: flattened generator
    """
    for element in nest:
        if isinstance(element, Iterable) and not isinstance(element, (str, bytes)):
            yield from flatten(element)
        else:
            yield element


def split_clean_paragraphs(iterrow: Iterable) -> list:
    """split_clean_paragraphs splits an article into paragraphs and removes additional newlines and spaces.
        Args:
            iterrow (Iterable): a pandas row object
        Returns:
            list: list of cleaned generated paragraphs
        """
    if iterrow.article_word_count <= 250:
        return [iterrow.content]
    splitted_paragraphs = re.split(r"\.\s*\n\s*\n", iterrow.content)
    splitted_paragraphs = [re.split(r"\n\n", i) if len(
        i.split()) > 1000 else i for i in splitted_paragraphs]
    return list(set(list(flatten(splitted_paragraphs))))


def split_clean_sentences(doc: spacy.tokens.doc.Doc, brand: str, brand_2: str, require_verb: bool = True) -> list:
    """split_clean_sentences breaks a document into sentences and cleans the text for faster processings.
    Args:
        doc (spacy.tokens.doc.Doc): spacy language document.
        brand (str): brand name.
        brand_2 (str): simplified brand name.
        require_verb (bool, optional): if true sentences without 'VERB' pos are removed. Defaults to True.
    Returns:
        list: list of clean sentences (spacy span objects).
    """
    if brand_2.split('-'):
        brand_2 = brand_2.split('-')[-1]
    elif len(brand_2.split()) > 1:
        brand_2 = brand_2.split()[0]
    clean_sentences = [sent for sent in doc.sents if bool(
        re.search('[a-z0-9]', sent.text, re.IGNORECASE))]
    clean_sentences = [
        sent for sent in clean_sentences
        if len(re.sub('[^A-Za-z0-9]+', ' ', re.sub(' +', ' ', sent.text).replace(
            ' |', ' ').replace(':::', ' ').replace('\n', ' ').strip()).split()) > 1
    ]
    if require_verb:
        clean_sentences = [sent for sent in clean_sentences if 'VERB' in [
            token.pos_ for token in sent]]
    clean_sentences = [
        sent for sent in clean_sentences if brand.lower() in
        sent.text.lower() or brand_2.lower() in sent.text.lower()
    ]
    return clean_sentences


def compute_relevancy_score_for_content(sentence: spacy.tokens.span.Span, brand: str, brand_2: str) -> float:
    # sourcery skip: remove-redundant-if
    """compute_relevancy_score_for_content assigns a relevancy score based on the grammatical importance of brank in article content.
    Args:
        sentence (spacy.tokens.span.Span): spacy span object (doc.sents).
        brand (str): brand name.
        brand_2 (str): simplified brand name.
    Returns:
        float: relevancy score.
    """
    if brand_2.split('-'):
        brand_2 = brand_2.split('-')[-1]
    if brand_2.split("'s"):
        brand_2 = brand_2.split("'s")[0]
    if len(brand_2.split()) > 1:
        brand_2 = brand_2.split()[0]

    relevancy_score = 0
    for token in sentence:
        if any(i.lower() in token.text.lower() for i in ['.com', '•', '@']):
            condition = brand_2.lower() in token.text.lower()
        else:
            condition = brand_2.lower() == token.text.lower()
        if token.text.lower() == brand.lower() or condition:
            brand_dep = token.dep_
            brand_ancestor_dep = [a.dep_ for a in token.ancestors]
            brand_children_dep = [c.dep_ for c in token.children]
            subj_idx_ancestor = [idx for idx, dep in enumerate(
                brand_ancestor_dep) if 'subj' in dep]
            subj_idx_children = [idx for idx, dep in enumerate(
                brand_children_dep) if 'subj' in dep]
            # subject rules
            if 'subj' in brand_dep:
                # print('subj rule success')
                relevancy_score += 1
            # dep tree rule
            elif any(i < 2 for i in subj_idx_ancestor) or any(i < 2 for i in subj_idx_children):
                # print('dep tree rule success')
                relevancy_score += 1
            elif any(i < 4 for i in subj_idx_ancestor) or any(i < 4 for i in subj_idx_children):
                # print('dep tree rule success')
                relevancy_score += 0.5
            # conjunction rules
            elif brand_dep == 'conj':
                # print('conj rule success')
                conjunct_with_dep = [
                    (idx, c, c.dep_)
                    for idx, c in enumerate(token.conjuncts)
                ]
                if any('subj' in i for i in [c[-1] for c in conjunct_with_dep]):
                    # print('conj-subj rule success')
                    relevancy_score += 1
                elif any('dobj' in i for i in [c[-1] for c in conjunct_with_dep]):
                    # print('conj-dobj rule success')
                    relevancy_score += 0.5
                elif any('appos' in i for i in [c[-1] for c in conjunct_with_dep]):
                    # print('conj-appos rule success')
                    relevancy_score += 0.25
                else:
                    # print('conj other rule success')
                    relevancy_score += 0.25
            # compound rules
            elif brand_dep == 'compound':
                # print('compound rule success')
                if any(i <= 3 for i in subj_idx_ancestor) or any(i <= 2 for i in subj_idx_children):
                    relevancy_score += 1
                # elif len(subj_idx_ancestor) != 0 or len(subj_idx_ancestor) != 0:
                #     relevancy_score += 0.25
                elif any(i == 'dobj' for i in brand_ancestor_dep[:2]):
                    relevancy_score += 0.5
                else:
                    relevancy_score += 0.25
            # appositional rules
            elif brand_dep == 'appos':
                # print('appos rule success')
                if any(i < 2 for i in subj_idx_children) or any(i < 2 for i in subj_idx_ancestor):
                    relevancy_score += 1
                # elif len(subj_idx_children) != 0:
                #     relevancy_score += 0.5
                elif brand_ancestor_dep[0] == 'dobj':
                    relevancy_score += 0.5
                else:
                    relevancy_score += 0.25
            # posseive rules
            elif brand_dep == 'poss':
                # print('poss rule success')
                if any(i < 2 for i in subj_idx_ancestor):
                    relevancy_score += 1
                else:
                    relevancy_score += 0.25
            # direct object rules
            elif brand_dep == 'dobj':
                # print('dobj rule success')
                relevancy_score += 0.5
            # prepositional object rules
            elif brand_dep == 'pobj':
                # print('pobj rule success')
                relevancy_score += 0.25
            else:
                relevancy_score += 0.25
    if brand_2.lower() == 'Johnson'.lower():
        relevancy_score /= 2
    return relevancy_score


def compute_relevancy_score_for_title(title: str, brand: str, brand_2: str) -> float:
    """compute_relevancy_score_for_title assigns a relevancy score based on the grammatical importance of brank in article title.
    Args:
        title (str): title of the article.
        brand (str): brand name.
        brand_2 (str): simplified brand name.
    Returns:
        float: relevancy score
    """
    title_doc = nlp_lg(title)
    sentences = split_clean_sentences(
        title_doc, brand, brand_2, require_verb=False)
    brand_dep = [
        (
            token.dep_,
            [
                a.dep_ for a in token.ancestors
            ][:2],
            [
                c.dep_ for c in token.children
            ][:2],
            [
                c.dep_ for c in token.conjuncts
            ]
        ) for sentence in sentences
        for token in sentence if token.text.lower() in
        [brand.lower(), brand_2.lower()]]

    brand_dep = list(flatten(brand_dep))

    # return sentences, brand_dep, relevancy_score
    return 2 if any('subj' in i for i in brand_dep) else 1



"""Module for package setup."""
from setuptools import find_packages, setup

setup(
    # Needed to silence warnings (and to be a worthwhile package)
    name='dxi_nlp',
    url='https://github.com/Edelman-DxI/dxi-nlp.git',
    author='Amit Prusty',
    author_email='amit.prusty@edelmandxi.com',
    # Needed to actually package something
    packages=find_packages(),
    namespace_packages=['dxi_nlp'],
    # Needed for dependencies
    install_requires=[
        'numpy', 'pandas', 'spacy',
        'boto3', 's3fs', 'matplotlib', 'tqdm',
        'pyarrow', 'labelbox',
    ],
    # *strongly* suggested for in-house use
    version='0.1.4',
    # The license can be anything you like
    license='Private',
    description='',
    # We will also need a readme eventually (there will be a warning)
    long_description=open('README.md').read(),
)
